from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99
# from a_WState import WState
# import moldb as db
from ...explorer.gui.gui_convmol.a_WMolecularConstants import *
from ... import hapi
import os
import datetime
from collections import OrderedDict
import f311.filetypes as ft
import f311.pyfant as pf
import f311.explorer as ex

__all__ = ["XConvMol"]


class _DataSource(a99.AttrsPart):
    """Represents a data source for molecular lines"""

    def __init__(self, name):
        a99.AttrsPart.__init__(self)
        self.name = name
        self.widget = None

    def __repr__(self):
        return "{}('{}')".format(self.__class__.__name__, self.name)



# This defines the order of the panels
_NAMES = ["HITRAN", "VALD3", "Kurucz", "TurboSpectrum", ]
_SOURCES = OrderedDict([[name, _DataSource(name)] for name in _NAMES])


class _WSource(a99.WBase):
    """Lists sources for molecular lines data"""

    @property
    def index(self):
        return self._get_index()

    @index.setter
    def index(self, x):
        self._buttons[x].setChecked(True)

    @property
    def source(self):
        """Returns _DataSource object or None"""
        i = self._get_index()
        if i < 0:
            return None
        return _SOURCES[_NAMES[i]]

    index_changed = pyqtSignal()

    def __init__(self, *args):
        a99.WBase.__init__(self, *args)

        self._buttons = []
        self._last_index = -1

        lw = QVBoxLayout()
        self.setLayout(lw)

        for ds in _SOURCES.values():
            b = self.keep_ref(QRadioButton(ds.name))
            b.clicked.connect(self._button_clicked)
            self._buttons.append(b)
            lw.addWidget(b)

    def _button_clicked(self):
        i = self._get_index()
        if i != self._last_index:
            self._last_index = i
            self.index_changed.emit()


    def _get_index(self):
        for i, b in enumerate(self._buttons):
            assert isinstance(b, QRadioButton)
            if b.isChecked():
                return i
        return -1


class _WSelectSaveFile(a99.WBase):
    @property
    def value(self):
        return self._get_value()

    @value.setter
    def value(self, x):
        self.edit.setText(x)

    # # Emitted whenever the valu property changes **to a valid value**
    wants_auto = pyqtSignal()

    def __init__(self, *args):
        a99.WBase.__init__(self, *args)

        self._last_value = None

        self._type = None
        self.dialog_title = "Save output as"
        self.dialog_path = "."
        self.dialog_wild = "*.*;;*.dat"

        lw = QHBoxLayout()
        self.setLayout(lw)

        t = self.label = self.keep_ref(QLabel("Save output as"))
        lw.addWidget(t)

        e = self.edit = QLineEdit()
        e.textChanged.connect(self.edit_changed)
        t.setBuddy(e)
        lw.addWidget(e)
        # e.setReadOnly(True)

        b = self.button_auto = QToolButton()
        lw.addWidget(b)
        b.clicked.connect(self.wants_auto)
        b.setIcon(a99.get_icon("leaf-plant"))
        b.setToolTip("Make up file name")
        b.setFixedWidth(30)

        b = self.button = QToolButton()
        lw.addWidget(b)
        b.clicked.connect(self.on_button_clicked)
        b.setIcon(a99.get_icon("document-save"))
        b.setToolTip("Choose file name to save as")
        b.setFixedWidth(30)

        # Forces paint red if invalid
        self.edit_changed()

    def on_button_clicked(self, _):
        self._on_button_clicked()

    def edit_changed(self):
        flag_valid = self.validate()
        a99.style_widget_valid(self.edit, not flag_valid)
        # if flag_valid:
        #     self._wanna_emit()

    def validate(self):
        """Returns True/False whether value is valid, i.e., existing file or directory"""
        value = self._get_value().strip()
        return len(value) > 0 and not os.path.isdir(value)

    def _on_button_clicked(self):
        path_ = self.edit.text().strip()
        if len(path_) == 0:
            path_ = self.dialog_path
        res = QFileDialog.getSaveFileName(self, self.dialog_title, path_, self.dialog_wild)[0]
        if res:
            # res = res[0]
            self.edit.setText(res)
            self.dialog_path = res

    # def _wanna_emit(self):
    #     value_now = self._get_value()
    #     if value_now != self._last_value:
    #         self._last_value = value_now
    #         self.valueChanged.emit()

    def _get_value(self):
        return self.edit.text().strip()


class _WHitranPanel(a99.WBase):

    @property
    def data(self):
        """
        Returns value in hapi.LOCAL_TABLE_CACHE, or None. This variable is a dictionary

        See hapi.py for structure of this variable
        """
        idx = self.tableWidget.currentRow()
        if idx < 0:
            return None
        return hapi.LOCAL_TABLE_CACHE[self.tableWidget.item(idx, 0).text()]

    def __init__(self, *args):
        a99.WBase.__init__(self, *args)

        self._flag_populating = False

        lw = QVBoxLayout()
        self.setLayout(lw)

        lw.addWidget(self.keep_ref(QLabel("HITRAN")))


        lg = QGridLayout()
        lw.addLayout(lg)

        a = self.keep_ref(QLabel("HITRAN 'data cache' directory"))
        w = self.w_dir = a99.WSelectDir(self.parent_form)
        a.setBuddy(w)
        w.valueChanged.connect(self.dir_changed)
        lg.addWidget(a, 0, 0)
        lg.addWidget(w, 0, 1)

        a = self.tableWidget = QTableWidget()
        lw.addWidget(a)
        a.setSelectionMode(QAbstractItemView.SingleSelection)
        a.setSelectionBehavior(QAbstractItemView.SelectRows)
        a.setEditTriggers(QTableWidget.NoEditTriggers)
        a.setFont(a99.MONO_FONT)
        a.setAlternatingRowColors(True)
        a.currentCellChanged.connect(self.on_tableWidget_currentCellChanged)

        # forces populate table with 'Python HITRAN API data cache' in local directory
        self.dir_changed()

    def on_tableWidget_currentCellChanged(self, curx, cury, prevx, prevy):
        pass

    def dir_changed(self):
        self._populate()


    def _populate(self):
        self._flag_populating = True
        try:
            # Changed HAPI working directory
            hapi.VARIABLES['BACKEND_DATABASE_NAME'] = self.w_dir.value
            # Loads all molecular lines data to memory
            hapi.loadCache()

            # Discounts "sampletab" table from HAPI cache, hence the "-1" below
            nr, nc = len(hapi.LOCAL_TABLE_CACHE)-1, 2
            t = self.tableWidget
            a99.reset_table_widget(t, nr, nc)
            t.setHorizontalHeaderLabels(["Table filename (.data & .header)", "Number of spectral lines"])

            i = 0
            for h, (name, data) in enumerate(hapi.LOCAL_TABLE_CACHE.items()):
                if name == "sampletab":
                    continue

                header = data["header"]

                item = QTableWidgetItem(name)
                t.setItem(i, 0, item)

                item = QTableWidgetItem(str(header["number_of_rows"]))
                t.setItem(i, 1, item)

                i += 1

            t.resizeColumnsToContents()

            # self._data = rows
            #
            # if restore_mode == "formula":
            #     if curr_row:
            #         self._find_formula(curr_row["formula"])
            # elif restore_mode == "index":
            #     if -1 < curr_idx < t.rowCount():
            #         t.setCurrentCell(curr_idx, 0)
        finally:
            self._flag_populating = False
            # self._wanna_emit_id_changed()


class _WVald3Panel(a99.WBase):
    """
    This panel allows to load a Vald3 file and browse through its species (molecules only)

    The goal is to choose one molecule
    """

    @property
    def data(self):
        """
        Returns FileVald3 with a single species"""

        idx = self.tableWidget.currentRow()
        if idx < 0:
            return None

        f = ft.FileVald3()
        f.speciess = [self._f.speciess[idx]]
        return f

    @property
    def is_molecule(self):
        data = self.data
        return data is not None and data.speciess[0].formula not in a99.symbols

    def __init__(self, *args):
        a99.WBase.__init__(self, *args)

        self._flag_populating = False
        self._f = None  # FileVald3

        lw = QVBoxLayout()
        self.setLayout(lw)

        lw.addWidget(self.keep_ref(QLabel("VALD3")))

        lg = QGridLayout()
        lw.addLayout(lg)

        a = self.keep_ref(QLabel("VALD3 extended-format file"))
        w = self.w_file = a99.WSelectFile(self.parent_form)
        a.setBuddy(w)
        w.valueChanged.connect(self.file_changed)
        lw.addWidget(w)
        lg.addWidget(a, 0, 0)
        lg.addWidget(w, 0, 1)


        a = self.tableWidget = QTableWidget()
        lw.addWidget(a)
        a.setSelectionMode(QAbstractItemView.SingleSelection)
        a.setSelectionBehavior(QAbstractItemView.SelectRows)
        a.setEditTriggers(QTableWidget.NoEditTriggers)
        a.setFont(a99.MONO_FONT)
        a.setAlternatingRowColors(True)
        a.currentCellChanged.connect(self.on_tableWidget_currentCellChanged)

        l = self.label_warning = QLabel()
        l.setStyleSheet("QLabel {{color: {}}}".format(a99.COLOR_WARNING))
        lw.addWidget(l)

        # forces populate table with 'Python HITRAN API data cache' in local directory
        # self.file_changed()

    def on_tableWidget_currentCellChanged(self, curx, cury, prevx, prevy):
        self.label_warning.setText("Need a molecule, not atom"
                                   if self.data is not None and not self.is_molecule else "")

    def file_changed(self):
        self._populate()


    def _populate(self):
        self._flag_populating = True
        try:
            f = self._f = ft.FileVald3()
            f.load(self.w_file.value)

            nr, nc = len(f), 3
            t = self.tableWidget
            a99.reset_table_widget(t, nr, nc)
            t.setHorizontalHeaderLabels(["VALD3 species", "Number of spectral lines", "Atom/Molecule"])

            for i, species in enumerate(f):
                item = QTableWidgetItem(str(species))
                t.setItem(i, 0, item)
                item = QTableWidgetItem(str(len(species)))
                t.setItem(i, 1, item)
                item = QTableWidgetItem("Atom" if species.formula in a99.symbols else "Molecule")
                t.setItem(i, 2, item)

            t.resizeColumnsToContents()

        except Exception as e:
            self._f = None
            self.add_log_error("Error reading contents of file '{}': '{}'".format(self.w_file.value, a99.str_exc(e)), True)
            raise

        else:
            self.clear_log()

        finally:
            self._flag_populating = False
            # self._wanna_emit_id_changed()

class _WKuruczPanel(a99.WBase):
    """
    This panel allows to load a Kurucz molecular lines file
    """

    @property
    def data(self):
        """Returns FileKuruczMoleculeBase or None"""
        return self._f

    @property
    def flag_hlf(self):
        return self.checkbox_hlf.isChecked()

    @property
    def flag_normhlf(self):
        return self.checkbox_normhlf.isChecked()

    @property
    def flag_fcf(self):
        return self.checkbox_fcf.isChecked()

    @property
    def flag_spinl(self):
        return self.checkbox_spinl.isChecked()

    @property
    def iso(self):
        idx = self.combobox_isotope.currentIndex()
        if idx == 0:
            return None
        return self._isotopes[idx-1]

    def __init__(self, *args):
        a99.WBase.__init__(self, *args)

        self._f = None  # FileKuruczMoleculeBase
        # list of integers, filled when file is loaded
        self._isotopes = []

        lw = QVBoxLayout()
        self.setLayout(lw)

        lw.addWidget(self.keep_ref(QLabel("Kurucz")))

        lg = QGridLayout()
        lw.addLayout(lg)

        i_row = 0
        a = self.keep_ref(QLabel("Kurucz molecular lines file"))
        w = self.w_file = a99.WSelectFile(self.parent_form)
        w.valueChanged.connect(self.file_changed)
        lg.addWidget(a, i_row, 0)
        lg.addWidget(w, i_row, 1)
        i_row += 1

        a = self.keep_ref(QLabel("Isotope"))
        w = self.combobox_isotope = QComboBox()
        # w.valueChanged.connect(self.file_changed)
        lg.addWidget(a, i_row, 0)
        lg.addWidget(w, i_row, 1)
        i_row += 1

        a = self.keep_ref(QLabel("Calculate 'gf' based on\n"
                                 "Hönl-London factors (HLFs)"))
        w = self.checkbox_hlf = QCheckBox()
        w.setToolTip("If selected, will ignore 'loggf' from Kurucz file and\n"
                     "calculate 'gf' using Hönl-London factors formulas taken from Kovacz (1969)")
        lg.addWidget(a, i_row, 0)
        lg.addWidget(w, i_row, 1)
        i_row += 1

        a = self.keep_ref(QLabel("Apply normalization factor\n"
                                 "for HLFs to add up to 1 (for fixed J)"))
        w = self.checkbox_normhlf = QCheckBox()
        w.setToolTip("If selected, calculated 'gf's will be multiplied by\n"
                     "2 / ((2 * J2l + 1) * (2 * S + 1)*(2 - DELTAK))")
        lg.addWidget(a, i_row, 0)
        lg.addWidget(w, i_row, 1)
        i_row += 1

        a = self.keep_ref(QLabel("Multiply calculated 'gf' by\n"
                                 "Franck-Condon factor"))
        w = self.checkbox_fcf = QCheckBox()
        w.setToolTip("If selected, incorporates internally calculated Franck-Condon factor"
                     "into the calculated 'gf'")
        lg.addWidget(a, i_row, 0)
        lg.addWidget(w, i_row, 1)
        i_row += 1

        a = self.keep_ref(QLabel("Use spin' for branch determination\n"
                                 "(spin'' is always used)"))
        w = self.checkbox_spinl = QCheckBox()
        w.setToolTip("If you tick this box, branches P12, P21, Q12, Q21, R21, R12 (i.e., with two numbers) become possible")
        lg.addWidget(a, i_row, 0)
        lg.addWidget(w, i_row, 1)
        i_row += 1


        lw.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))

    def file_changed(self):
        cb = self.combobox_isotope
        try:
            f = self._f = ft.load_kurucz_mol(self.w_file.value)

            self.update_gui_iso(cb, f)
        except Exception as e:
            self._f = None
            self._isotopes = []
            cb.clear()
            msg = "Error reading contents of file '{}': '{}'".format(self.w_file.value,
                                                                              a99.str_exc(e))
            self.add_log_error(msg, True)
            a99.get_python_logger().exception(msg)

    def update_gui_iso(self, cb, f):
        """Updates Isotope combobox"""
        cb = self.combobox_isotope
        cb.clear()
        if f.__class__ == ft.FileKuruczMoleculeOld:
            cb.addItem("(all (file is old-format))")
        else:
            self._isotopes = list(set([line.iso for line in f]))
            self._isotopes.sort()
            cb.addItem("(all)")
            cb.addItems([str(x) for x in self._isotopes])


class _WTurboSpectrumPanel(a99.WBase):
    def __init__(self, *args):
        a99.WBase.__init__(self, *args)

        lw = QVBoxLayout()
        self.setLayout(lw)

        lw.addWidget(self.keep_ref(QLabel("TurboSpectrum")))



class XConvMol(ex.XFileMainWindow):
    def __init__(self, parent=None, fileobj=None):
        ex.XFileMainWindow.__init__(self, parent)

        # # Synchronized sequences
        _VVV = ft.FileMolDB.description
        self.tab_texts =  ["{} (Alt+&1)".format(_VVV), "Conversion (Alt+&2)", "Log (Alt+&3)"]
        self.flags_changed = [False, False]
        self.save_as_texts = ["Save %s as..." % _VVV, None, None]
        self.open_texts = ["Load %s" % _VVV, None, None]
        self.clss = [ft.FileMolDB, None, None]  # save class
        self.clsss = [(ft.FileMolDB,), None, None]  # accepted load classes
        self.wilds = ["*.sqlite", None, None]  # e.g. '*.fits'
        self.editors = [ex.NullEditor(), ex.NullEditor(), ex.NullEditor()]  # editor widgets, must comply ...
        tw0 = self.tabWidget
        tw0.setTabText(1, self.tab_texts[2])


        lv = self.keep_ref(QVBoxLayout(self.gotting))
        me = self.w_moldb = WMolecularConstants(self)
        lv.addWidget(me)
        me.changed.connect(self._on_changed)
        self.editors[0] = me


        # # Second tab: files

        w = self.keep_ref(QWidget())
        tw0.insertTab(1, w, self.tab_texts[1])


        # ## Vertical layout: source and destination stacked
        lsd = self.keep_ref(QVBoxLayout(w))

        # ### Horizontal layout: sources radio buttons, source-specific setup area

        lh = self.keep_ref(QHBoxLayout())
        lsd.addLayout(lh)

        # #### Vertical layout: source radio group box

        lss = QVBoxLayout()
        lh.addLayout(lss)

        # ##### Source radio buttons
        lss.addWidget(self.keep_ref(QLabel("<b>Source</b>")))
        w = self.w_source = _WSource(self)
        w.index_changed.connect(self.source_changed)
        lss.addWidget(w)
        lss.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))


        # #### Adds configuration panels for various sources
        # Only one panel should be visible at a time
        # **Note** The order here doesn't matter
        panels = {}
        panels["HITRAN"] = self.w_hitran = _WHitranPanel(self)
        panels["VALD3"] = self.w_vald3 = _WVald3Panel(self)
        panels["TurboSpectrum"] = self.w_turbo = _WTurboSpectrumPanel(self)
        panels["Kurucz"] = self.w_kurucz = _WKuruczPanel(self)
        for name in _NAMES:
            p = panels[name]
            ds = _SOURCES[name]
            ds.widget = p
            lh.addWidget(p)

        # ### Output file specification

        w0 = self.w_out = _WSelectSaveFile(self)
        w0.wants_auto.connect(self.wants_auto)
        lsd.addWidget(w0)

        # ### "Convert" button

        lmn = QHBoxLayout()
        lsd.addLayout(lmn)
        lmn.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))
        b = self.button_convert = QPushButton("&Run Conversion")
        b.clicked.connect(self.convert_clicked)
        lmn.addWidget(b)
        b = self.button_convert = QPushButton("&Open result in mled.py")
        b.clicked.connect(self.open_mol_clicked)
        lmn.addWidget(b)

        # # Final adjustments

        tw0.setCurrentIndex(0)
        # Forces only one of the source panels to visible
        self.w_source.index = 0
        self.source_changed()
        self.setWindowTitle("(to) PFANT Molecular Lines Converter")

        a99.nerdify(self)

        if fileobj is not None:
            self.load(fileobj)

        self.installEventFilter(self)

    def eventFilter(self, obj_focused, event):
        if event.type() == QEvent.KeyPress:
            # To help my debugging: Configures for Kurucz OH conversion
            if event.key() == Qt.Key_F12:
                self.w_moldb.combobox_select_molecule.setCurrentIndex(7)
                self.w_moldb.combobox_select_pfantmol.setCurrentIndex(1)
                self.w_moldb.combobox_select_state.setCurrentIndex(6)
                self.w_moldb.combobox_select_system.setCurrentIndex(0)
                self.w_moldb.None_to_zero()
                self.w_source._buttons[2].setChecked(True)
                self.source_changed()
                self.w_kurucz.w_file.edit.setText("ohaxupdate.asc")
                self.tabWidget.setCurrentIndex(1)
        return False

    def wants_auto(self):
        name = _NAMES[self.w_source.index]
        filename = None
        if name == "HITRAN":
            lines = self.w_hitran.data
            if lines:
                filename = "{}.dat".format(lines["header"]["table_name"])

        if filename is None:
            # Default
            filename = a99.new_filename("mol", "dat")
        self.w_out.value = filename


    def source_changed(self):
        idx = self.w_source.index
        for i, ds in enumerate(_SOURCES.values()):
            ds.widget.setVisible(i == idx)
            # print("Widget", ds.widget, "is visible?", ds.widget.isVisible())

    def convert_clicked(self):
        from f311 import convmol as cm
        try:
            # # Extraction of data from GUI, and their
            name = self.w_source.source.name
            # This is a merge of fields table 'molecule' and 'pfantmol' in a FileMolDB database
            mol_consts = self.w_moldb.constants
            fcfs = self.w_moldb.fcfs
            filename = self.w_out.value

            # Validation of data
            errors = []
            if self.w_moldb.id_molecule is None:
                errors.append("Molecule not selected")
            elif any([x is None for x in mol_consts.values()]):
                s_none = ", ".join(["'{}'".format(key) for key, value in mol_consts.items() if value is None])
                errors.append("There are empty molecular constants: {}".format(s_none))
            if not self.w_out.validate():
                errors.append("Output filename is invalid")

            lines, sols_calculator = None, None
            if len(errors) == 0:
                # Source-dependant calculation of "sets of lines"
                if name == "HITRAN":
                    errors.append("HITRAN conversion not implemented yet!")
                    if False:
                        lines = self.w_hitran.data

                        if lines is None:
                            errors.append("HITRAN table not selected")
                        else:
                            conv = cm.ConvHitran()
                elif name == "VALD3":
                    errors.append("VALD3 conversion not implemented yet!")
                    if False:
                        if not self.w_vald3.is_molecule:
                            errors.append("Need a VALD3 molecule")
                        else:
                            lines = self.w_vald3.data
                            conv = cm.ConvVald3()
                elif name == "Kurucz":
                    w = self.w_kurucz

                    if w.flag_fcf and fcfs is None:
                        errors.append("Cannot multiply gf's by Franck-Condon Factors, as these are not available in molecular configuration")
                    else:

                        conv = cm.ConvKurucz(flag_hlf=w.flag_hlf, flag_normhlf=w.flag_normhlf,
                                             flag_fcf=w.flag_fcf, flag_spinl=w.flag_spinl, iso=w.iso)
                        lines = w.data
                else:
                    a99.show_message("{}-to-PFANT conversion not implemented yet, sorry".
                                    format(name))
                    return

            if len(errors) == 0:
                # Finally the conversion to PFANT molecular lines file
                conv.mol_consts = mol_consts
                conv.fcfs = fcfs
                f, log = conv.make_file_molecules(lines)
                ne = len(log.errors)
                if ne > 0:
                    self.add_log("Error messages:")
                    self.add_log("\n".join(log.errors))
                    self.add_log("{} error message{}".format(ne, "" if ne == 1 else "s"))

                if log.flag_ok:
                    f.save_as(filename)
                    if log.num_lines_skipped > 0:
                        self.add_log(
                            "Lines filtered out: {}".format(log.num_lines_skipped))
                        self.add_log("    Reasons:")
                        kv = list(log.skip_reasons.items())
                        kv.sort(key=lambda x: x[0])
                        for key, value in kv:
                            self.add_log("      - {}: {}".format(key, value))
                    ne = log.num_lines-log.num_lines_skipped-f.num_lines
                    if ne > 0:
                        self.add_log(
                            "Lines not converted because of error: {}".format(ne))
                    self.add_log(
                        "Lines converted: {}/{}".format(f.num_lines, log.num_lines))

                    self.add_log("File '{}' generated successfully".format(filename))
                else:
                    self.add_log_error("Conversion was not possible")
            else:
                self.add_log_error("Cannot convert:\n  - " + ("\n  - ".join(errors)), True)

        except Exception as e:
            a99.get_python_logger().exception("Conversion failed")
            self.add_log_error("Conversion failed: {}".format(a99.str_exc(e)), True)


    def open_mol_clicked(self):
        filename = self.w_out.value
        if len(filename) > 0:
            f = ft.FileMolecules()
            f.load(filename)
            vis = ex.VisMolecules()
            vis.use(f)



"""
Multiplicity (Chemistry)

https://en.wikipedia.org/wiki/Multiplicity_(chemistry)

  - Formulas to calculate the Hönl-London factors
  - Conversion from quantum information to branch label such as P/Q/R/P1/Q1 etc.

Note: This package will preserve the namespace hierarchy because there are repeated method names, e.g.,
`quanta_to_branch()`
"""


#
# _J2L_MAP = {"P": -1, "Q": 0, "R": 1}
# def branch_to_J2l(br):
#     """Converts P/Q/R to -1/0/+1
#
#     Tolerant to strings such as "P1", "P12" etc. (i.e., takes only the first character)"""
#     return _J2L_MAP[br[0]]


from ..molconsts import MolConsts

class honllondon_dict(dict):
    """
    Subclass of dict to deal with keys as (vl, v2l, J, branch)

    References:
        Based on solution by Rochen Ritzel at
        https://stackoverflow.com/questions/2912231/is-there-a-clever-way-to-pass-the-key-to-defaultdicts-default-factory
    """

    def __init__(self, mol_consts):
        if not isinstance(mol_consts, MolConsts):
            raise TypeError("mol_consts must be a MolConsts")
        dict.__init__(self)
        self._mol_consts = mol_consts

    def __missing__(self, key):
        self._populate_with_key(key)
        return self[key]

    def _populate_with_key(self, key):
        """Must be inherited and populate self with all branches for given (vl, v2l, J)"""
        raise NotImplementedError()


from . import singlet
from . import doublet
from . import triplet



from .calc_qgbd import *
from .convlog import *
import datetime

__all__ = ["Conv"]



class Conv(object):
    # TODO there is a good chance that flag_hlf, flag_fcf, etc. will be moved to this class
    """Molecular lines converter base class

    Args:
        qgbd_calculator: callable that can calculate "qv", "gv", "bv", "dv",
                         Default: calc_qbdg_tio_like

        mol_consts: a dict-like object combining field values from tables 'molecule', 'state',
                    'pfantmol', and 'system' from a FileMolDB database

        fcfs: Franck-Condon Factors (dictionary of floats indexed by (vl, v2l))
    """

    def __init__(self, qgbd_calculator=None, mol_consts=None, fcfs=None):
        self.qgbd_calculator = qgbd_calculator if qgbd_calculator else calc_qgbd_tio_like
        self.mol_consts = mol_consts
        self.fcfs = fcfs

    def make_file_molecules(self, lines):
        """
        Builds a FileMolecules object

        Args:
            lines: in most cases, a DataFile instance, but may be anything, since the molecular
                   lines data extraction is performed by a particular descendant class.
                   For example, the HITRAN converter expects a structure from the HAPI

        Returns:
            f, log: FileMolecules, MolConversionLog instances
        """
        from f311 import filetypes as ft

        # Runs specific conversor to SetOfLines
        sols, log = self._make_sols(lines)
        assert isinstance(sols, list)
        assert isinstance(log, MolConversionLog)

        sols.sort(key= lambda sol: sol.vl*1000+sol.v2l)


        mol = ft.mol_consts_to_molecule(self.mol_consts)
        mol.sol = sols
        f = ft.FileMolecules()
        now = datetime.datetime.now()
        f.titm = "Created by ftpyfant.make_file_molecules() @ {}".format(now.isoformat())
        f.molecules = [mol]
        return f, log

    # Must reimplement thig
    def _make_sols(self, lines):
        """Converts molecular lines into a list of SetOfLines object

        Args:
            lines: see make_file_molecules()

        Returns:
            sols, log: list of SefOfLines object, MolConversionLog object
        """
        raise NotImplementedError()

    def _calculate_qgbd(self, v2l):
        return self.qgbd_calculator(self.mol_consts, v2l)



"""
Hönl-London factors for triplets.

Only the case for Delta Lambda = +-1 implemented.
Formulas from Kovacs 1969 p136, p137, and formula 2.1.4-8

Original comments in Portuguese were quoted from ATMOS/wrk4/bruno/Mole/NH/sjnh.f

Routines in uppercase are adapted code from Fortran

- `YL`, `U*L` refer to the superior state
- `Y2l`, `U*2L` refer to the inferior state 

"""

from collections import OrderedDict
import a99
from . import doublet
from . import honllondon_dict
import math


__all__ = ["honllondon", "quanta_to_branch"]


quanta_to_branch = doublet.quanta_to_branch


class honllondon(honllondon_dict):
    def _populate_with_key(self, key):
        vl, v2l, J, branch = key
        cc = self._mol_consts

        S = cc["s"]
        DELTAK = cc["cro"]
        FE = cc["fe"]
        LAML = cc["from_spdf"]
        LAM2L = cc["to_spdf"]

        # Original comment:
        #
        #     cada banda possue valores distintos de Y' e Y''
        #     quando Y=A/B > 0, temos o termo normal, por isso
        #     nao se modifica as formulas abaixo

        # TODO link these names in the documentation, gui_info, caption, description, whatever

        LAML = cc["from_spdf"]
        LAM2L = cc["to_spdf"]
        AL = cc["statel_A"]
        AEL = cc["statel_alpha_e"]
        BEL = cc["statel_B_e"]
        A2L = cc["state2l_A"]
        AE2L = cc["state2l_alpha_e"]
        BE2L = cc["state2l_B_e"]

        BL = BEL - AEL * (vl + 0.5)
        B2L = BE2L - AE2L * (v2l + 0.5)

        YL = AL / BL
        Y2L = A2L / B2L

        # Original comment:
        #
        #     cálculo das contantes U1+, U1-, U3+, U3-,
        #     C1+, C2+, C3+, dependentes de J para o estado
        #     eletronico superior( A 3PI, lambda=1)
        #     formulas para casos intermediarios entre os casos A( Y>>J(J+1) )
        #     e B( Y<<J(J+1) ) de Hund
        #
        # Note: I (JT) changed some entity names for clarity and generality. For example:
        #
        #     - U1PLUSL was originally called U1MAA, ("MA" stood for "MAIS", "plus" in Portuguese;
        #       "A" stood the particular state for which the original NH code was developed and was replaced
        #       by "L", meaning "linha", to be consistent with a notation found in many places in the code)
        #     - "YL" was replaced by "YL", and "Y2L" was replaced by "Y2L" because, again, "L" and "2L"
        #       are used in other places to denote the superior and inferior levels, respectively

        U1PLUSL = lambda J: math.sqrt(LAML*LAML*YL*(YL-4)+4*J*J)+(LAML*(YL-2))
        U1MINUSL = lambda J: math.sqrt(LAML*LAML*YL*(YL-4)+4*J*J)-(LAML*(YL-2))
        U3PLUSL = lambda J: (LAML*LAML*YL*(YL-4)+4*(J+1)*(J+1))**.5+LAML*(YL-2)
        U3MINUSL = lambda J: ((LAML*LAML*YL*(YL-4)+4*(J+1)*(J+1))**0.5)-LAML*(YL-2)
        C1L = lambda J: LAML*LAML*YL*(YL-4)*(J-LAML+1)*(J+LAML)+2*(2*J+1)*(J-LAML)*(J+LAML)*J
        C2L = lambda J: LAML*LAML*YL*(YL-4)+4*J*(J+1)
        C3L = lambda J: LAML*LAML*YL*(YL-4)*(J-LAML)*(J+LAML+1)+2*(2*J+1)*(J-LAML+1)*(J+1)*(J+LAML+1)

        # Original comment:
        #
        #     idem para o estado eletronico inferior,
        #     X 3Sig( lambda=0)

        U1PLUS2L = lambda J: ((LAM2L*LAM2L*Y2L*(Y2L-4)+4*J*J))**.5+LAM2L*(Y2L-2)
        U1MINUS2L = lambda J: ((LAM2L*LAM2L*Y2L*(Y2L-4)+4*J*J))**.5-LAM2L*(Y2L-2)
        U3PLUS2L = lambda J: ((LAM2L*LAM2L*Y2L*(Y2L-4)+4*(J+1)*(J+1))**.5)+LAM2L*(Y2L-2)
        U3MINUS2L = lambda J: ((LAM2L*LAM2L*Y2L*(Y2L-4)+4*(J+1)*(J+1))**.5)-LAM2L*(Y2L-2)
        C12L = lambda J: LAM2L*LAM2L*Y2L*(Y2L-4)*(J-LAM2L+1)*(J+LAM2L)+2*(2*J+1)*(J-LAM2L)*J*(J+LAM2L)
        C22L = lambda J: LAM2L*LAM2L*Y2L*(Y2L-4)+4*J*(J+1)
        C32L = lambda J: LAM2L*LAM2L*Y2L*(Y2L-4)*(J-LAM2L)*(J+LAM2L+1)+2*(2*J+1)*(J-LAM2L+1)*(J+1)*(J+LAM2L+1)


        # Original comment:
        #
        #     usa-se o menor valor de Lambda nas formulas abaixo
        LMIN = min(LAML, LAM2L)


        _P1 = lambda J:\
            ((J-LMIN-1)*(J-LMIN))/(32*J*C1L(J-1)*C12L(J))* \
            (((J-LMIN+1)*(J+LMIN)*U1PLUSL(J-1)*U1PLUS2L(J)+
            (J-LMIN-2)*(J+LMIN+1)*U1MINUSL(J-1)*U1MINUS2L(J)+
            8*(J-LMIN-2)*(J-LMIN)*(J+LMIN)*(J+LMIN))**2)

        _Q1 = lambda J: \
            ((J-LMIN)*(J+LMIN+1)*(2*J+1))/(32*J*(J+1)*C1L(J)*C12L(J)) \
            *(((J-LMIN+1)*(J+LMIN)*U1PLUSL(J)*U1PLUS2L(J)+
            (J-LMIN-1)*(J+LMIN+2)*U1MINUSL(J)*U1MINUS2L(J)+
            8*(J-LMIN-1)*(J-LMIN)*(J+LMIN)*(J+LMIN+1))**2)

        _R1 = lambda J: \
            (((J+LMIN+1)*(J+LMIN+2))/(32*(J+1)*
            C1L(J+1)*C12L(J)))*(((J-LMIN+1)*(J+LMIN)*
            U1PLUSL(J+1)*U1PLUS2L(J)+(J-LMIN)*(J+LMIN+3)*U1MINUSL(J+1)*
            U1MINUS2L(J)+8*(J-LMIN)*(J-LMIN)*(J+LMIN)*(J+LMIN+2))**2)


        _P21 = lambda J: \
            (((J-LMIN-1)*(J-LMIN))/(4*J*C2L(J-1)*C12L(J)))* \
            (((J-LMIN+1)*(J+LMIN)*U1PLUS2L(J)-
            (J-LMIN-2)*(J+LMIN+1)*U1MINUS2L(J)-
            2*(LMIN+1)*(J-LMIN)*(J+LMIN)*(YL-2))**2)

        _Q21 = lambda J: \
            (((J-LMIN)*(J+LMIN+1)*(2*J+1))/(4*J*(J+1)*C2L(J)*C12L(J)))* \
            (((J-LMIN+1)*(J+LMIN)*U1PLUS2L(J)-
            (J-LMIN-1)*(J+LMIN+2)*U1MINUS2L(J)-
            2*(LMIN+1)*(J-LMIN)*(J+LMIN)*(YL-2))**2)

        _R21 = lambda J: \
            (((J+LMIN+1)*(J+LMIN+2))/(4*(J+1)*
            C2L(J+1)*C12L(J)))*(((J-LMIN+1)*(J+LMIN)*
            U1PLUS2L(J) - (J-LMIN)*(J+LMIN+3)*U1MINUS2L(J)
            - 2*(LMIN+1)*(J-LMIN)*(J+LMIN)*(YL-2))**2)

        _P31 = lambda J: \
            (((J-LMIN-1)*(J-LMIN))/(32*J*C3L(J-1)*C12L(J)))* \
            (((J-LMIN+1)*(J+LMIN)*U3MINUSL(J-1)*U1PLUS2L(J)+
            (J-LMIN-2)*(J+LMIN+1)*U3PLUSL(J-1)*U1MINUS2L(J)-
            8*(J-LMIN-1)*(J-LMIN)*(J+LMIN)*(J+LMIN+1))**2)

        _Q31 = lambda J: \
            (((J-LMIN)*(J+LMIN+1)*(2*J+1))/(32*J*(J+1)*C3L(J)*C12L(J))) \
            *(((J-LMIN+1)*(J+LMIN)*U3MINUSL(J)*U1PLUS2L(J)+
            (J-LMIN-1)*(J+LMIN+2)*U3PLUSL(J)*U1MINUS2L(J)-
            8*(J-LMIN)*(J-LMIN)*(J+LMIN)*(J+LMIN+2))**2)

        _R31 = lambda J: \
            (((J+LMIN+1)*(J+LMIN+2))/(32*(J+1)*
            C3L(J+1)*C12L(J)))*(((J-LMIN+1)*(J+LMIN)*
            U3MINUSL(J+1)*U1PLUS2L(J)+(J-LMIN)*(J+LMIN+3)*U3PLUSL(J+1)*
            U1MINUS2L(J)- 8*(J-LMIN)*(J-LMIN+1)*(J+LMIN)*(J+LMIN+3))**2)

        _P12 = lambda J: \
            (((J-LMIN-1)*(J-LMIN))/(4*J*C1L(J-1)*C22L(J)))* \
            (((J-LMIN+1)*(J+LMIN)*U1PLUSL(J-1)-
            (J-LMIN-2)*(J+LMIN+1)*U1MINUSL(J-1)-
            2*LMIN*(J-LMIN-2)*(J+LMIN)*(Y2L-2))**2)

        _Q12 = lambda J: \
            (((J-LMIN)*(J+LMIN+1)*(2*J+1))/(4*J*(J+1)*C1L(J)*C22L(J))) \
            *(((J-LMIN+1)*(J+LMIN)*U1PLUSL(J)-
            (J-LMIN-1)*(J+LMIN+2)*U1MINUSL(J)-
            2*LMIN*(J-LMIN-1)*(J+LMIN+1)*(Y2L-2))**2)

        _R12 = lambda J: \
            (((J+LMIN+1)*(J+LMIN+2))/(4*(J+1)*
            C1L(J+1)*C22L(J)))*(((J-LMIN+1)*(J+LMIN)*
            U1PLUSL(J+1) - (J-LMIN)*(J+LMIN+3)*U1MINUSL(J+1)
            - 2*LMIN*(J-LMIN)*(J+LMIN+2)*(Y2L-2))**2)

        _P2 = lambda J: \
            ((2*(J-LMIN-1)*(J-LMIN))/(J*C2L(J-1)*C22L(J)))* \
            ((.5*LMIN*(LMIN+1)*(YL-2)*(Y2L-2)+(J-LMIN+1)*(J+LMIN)+
            (J-LMIN-2)*(J+LMIN+1))**2)

        _Q2 = lambda J: \
            ((2*(J-LMIN)*(J+LMIN+1)*(2*J+1))/(J*(J+1)*C2L(J)*C22L(J)))* \
            ((.5*LMIN*(LMIN+1)*(YL-2)*(Y2L-2)+
            (J-LMIN+1)*(J+LMIN)+(J-LMIN-1)*(J+LMIN+2))**2)

        _R2 = lambda J: \
            ((2*(J+LMIN+1)*(J+LMIN+2))/((J+1)*
            C2L(J+1)*C22L(J)))*((.5*LMIN*(LMIN+1)*(YL-2)*(Y2L-2)+
            (J-LMIN+1)*(J+LMIN)+(J-LMIN)*(J+LMIN+3))**2)

        _P32 = lambda J: \
            (((J-LMIN-1)*(J-LMIN))/(4*J*C3L(J-1)*C22L(J)))* \
            (((J-LMIN+1)*(J+LMIN)*U3MINUSL(J-1)-
            (J-LMIN-2)*(J+LMIN+1)*U3PLUSL(J-1)+
            2*LMIN*(J-LMIN-1)*(J+LMIN+1)*(Y2L-2))**2)

        _Q32 = lambda J: \
            (((J-LMIN)*(J+LMIN+1)*(2*J+1))/(4*J*(J+1)*C3L(J)*C22L(J))) \
            *(((J-LMIN+1)*(J+LMIN)*U3MINUSL(J)-
            (J-LMIN-1)*(J+LMIN+2)*U3PLUSL(J)+
            2*LMIN*(J-LMIN)*(J+LMIN+2)*(Y2L-2))**2)

        _R32 = lambda J: \
            (((J+LMIN+1)*(J+LMIN+2))/(4*(J+1)*
            C3L(J+1)*C22L(J)))*(((J-LMIN+1)*(J+LMIN)*
            U3MINUSL(J+1) - (J-LMIN)*(J+LMIN+3)*U3PLUSL(J+1)
            + 2*LMIN*(J-LMIN+1)*(J+LMIN+3)*(Y2L-2))**2)

        _P13 = lambda J: \
            (((J-LMIN-1)*(J-LMIN))/(32*J*C1L(J-1)*
            C32L(J)))*(((J-LMIN+1)*(J+LMIN)*U1PLUSL(J-1)*
            U3MINUS2L(J)+ (J-LMIN-2)*(J+LMIN+1)*U1MINUSL(J-1)*U3PLUS2L(J)-
            8*(J-LMIN-2)*(J-LMIN+1)*(J+LMIN)*(J+LMIN+1))**2)

        _Q13 = lambda J: \
            (((J-LMIN)*(J+LMIN+1)*(2*J+1))/(32*J*(J+1)*C1L(J)*C32L(J)))* \
            (((J-LMIN+1)*(J+LMIN)*U1PLUSL(J)*U3MINUS2L(J)+
            (J-LMIN-1)*(J+LMIN+2)*U1MINUSL(J)*U3PLUS2L(J)-
            8*(J-LMIN-1)*(J-LMIN+1)*(J+LMIN+1)*(J+LMIN+1))**2)

        _R13 = lambda J: \
            (((J+LMIN+1)*(J+LMIN+2))/(32*(J+1)*
            C1L(J+1)*C32L(J)))*(((J-LMIN+1)*(J+LMIN)*U1PLUSL(J+1)
            *U3MINUS2L(J)+(J-LMIN)*(J+LMIN+3)*U1MINUSL(J+1)*U3PLUS2L(J)-
            8*(J-LMIN)*(J-LMIN+1)*(J+LMIN+1)*(J+LMIN+2))**2)

        _P23 = lambda J: \
            (((J-LMIN-1)*(J-LMIN))/(4*J*C2L(J-1)*C32L(J)))* \
            (((J-LMIN+1)*(J+LMIN)*U3MINUS2L(J)-
            (J-LMIN-2)*(J+LMIN+1)*U3PLUS2L(J)+
            2*(LMIN+1)*(J-LMIN+1)*(J+LMIN+1)*(YL-2))**2)

        _Q23 = lambda J: \
            (((J-LMIN)*(J+LMIN+1)*(2*J+1))/(4*J*(J+1)*C2L(J)*C32L(J))) \
            *(((J-LMIN+1)*(J+LMIN)*U3MINUS2L(J)-
            (J-LMIN-1)*(J+LMIN+2)*U3PLUS2L(J)+
            2*(LMIN+1)*(J-LMIN+1)*(J+LMIN+1)*(YL-2))**2)

        _R23 = lambda J: \
            (((J+LMIN+1)*(J+LMIN+2))/(4*(J+1)*
            C2L(J+1)*C32L(J)))*(((J-LMIN+1)*(J+LMIN)*
            U3MINUS2L(J) - (J-LMIN)*(J+LMIN+3)*U3PLUS2L(J)
            + 2*(LMIN+1)*(J-LMIN+1)*(J+LMIN+1)*(YL-2))**2)

        _P3 = lambda J: \
            (((J-LMIN-1)*(J-LMIN))/(32*J*C3L(J-1)*
            C32L(J)))*(((J-LMIN+1)*(J+LMIN)*U3MINUSL(J-1)*
            U3MINUS2L(J)+(J-LMIN-2)*(J+LMIN+1)*U3PLUSL(J-1)*U3PLUS2L(J)+
            8*(J-LMIN-1)*(J-LMIN+1)*(J+LMIN+1)*(J+LMIN+1))**2)

        _Q3 = lambda J: \
            (((J-LMIN)*(J+LMIN+1)*(2*J+1))/(32*J*(J+1)*C3L(J)*C32L(J)))* \
            (((J-LMIN+1)*(J+LMIN)*U3MINUSL(J)*U3MINUS2L(J)+
            (J-LMIN-1)*(J+LMIN+2)*U3PLUSL(J)*U3PLUS2L(J)+
            8*(J-LMIN)*(J-LMIN+1)*(J+LMIN+1)*(J+LMIN+2))**2)

        _R3 = lambda J: \
            (((J+LMIN+1)*(J+LMIN+2))/(32*(J+1)*
            C3L(J+1)*C32L(J)))*(((J-LMIN+1)*(J+LMIN)*U3MINUSL(J+1)
            *U3MINUS2L(J)+(J-LMIN)*(J+LMIN+3)*U3PLUSL(J+1)*U3PLUS2L(J)+
            8*(J-LMIN+1)*(J-LMIN+1)*(J+LMIN+1)*(J+LMIN+3))**2)



# NEXT: RESOLVE THE DELTA_LAMBDA AND STUFF THE OBJECT



# TODO cleanup
    # Original comment:
    #
    #     cálculo das contantes U1+, U1-, U3+, U3-,
    #     C1+, C2+, C3+, dependentes de J para o estado
    #     eletronico superior( A 3PI, lambda=1)
    #     formulas para casos intermediarios entre os casos A( Y>>J(J+1) )
    #     e B( Y<<J(J+1) ) de Hund
    #
    # Note: I (JT) changed some entity names for clarity and generality. For example:
    #
    #     - U1PLUSL was originally called U1MAA, ("MA" stood for "MAIS", "plus" in Portuguese;
    #       "A" stood the particular state for which the original NH code was developed and was replaced
    #       by "L", meaning "linha", to be consistent with a notation found in many places in the code)
    #     - "YL" was replaced by "YL", and "Y2L" was replaced by "Y2L" because, again, "L" and "2L"
    #       are used in other places to denote the superior and inferior levels, respectively

    # def U1PLUSL(J, LAML, YL):
    #    return math.sqrt(LAML*LAML*YL*(YL-4)+4*J*J)+(LAML*(YL-2))
    #
    # def U1MINUSL(J, LAML, YL):
    #     return math.sqrt(LAML*LAML*YL*(YL-4)+4*J*J)-(LAML*(YL-2))
    #
    # def U3PLUSL(J, LAML, YL):
    #     return (LAML*LAML*YL*(YL-4)+4*(J+1)*(J+1))**.5+LAML*(YL-2)
    #
    # def U3MINUSL(J, LAML, YL):
    #     return ((LAML*LAML*YL*(YL-4)+4*(J+1)*(J+1))**0.5)-LAML*(YL-2)
    #
    # def C1L(J, LAML, YL):
    #     return LAML*LAML*YL*(YL-4)*(J-LAML+1)*(J+LAML)+2*(2*J+1)*(J-LAML)*(J+LAML)*J
    #
    # def C2L(J, LAML, YL):
    #     return LAML*LAML*YL*(YL-4)+4*J*(J+1)
    #
    # def C3L(J, LAML, YL):
    #     return LAML*LAML*YL*(YL-4)*(J-LAML)*(J+LAML+1)+2*(2*J+1)*(J-LAML+1)*(J+1)*(J+LAML+1)
    #
    # # Original comment:
    # #
    # #     idem para o estado eletronico inferior,
    # #     X 3Sig( lambda=0)
    #
    # def U1PLUS2L(J, LAM2L, Y2L):
    #     return ((LAM2L*LAM2L*Y2L*(Y2L-4)+4*J*J))**.5+LAM2L*(Y2L-2)
    #
    # def U1MINUS2L(J, LAM2L, Y2L):
    #     return ((LAM2L*LAM2L*Y2L*(Y2L-4)+4*J*J))**.5-LAM2L*(Y2L-2)
    #
    # def U3PLUS2L(J, LAM2L, Y2L):
    #     return ((LAM2L*LAM2L*Y2L*(Y2L-4)+4*(J+1)*(J+1))**.5)+LAM2L*(Y2L-2)
    #
    # def U3MINUS2L(J, LAM2L, Y2L):
    #     return ((LAM2L*LAM2L*Y2L*(Y2L-4)+4*(J+1)*(J+1))**.5)-LAM2L*(Y2L-2)
    #
    # def C12L(J, LAM2L, Y2L):
    #     return LAM2L*LAM2L*Y2L*(Y2L-4)*(J-LAM2L+1)*(J+LAM2L)+2*(2*J+1)*(J-LAM2L)*J*(J+LAM2L)
    #
    # def C22L(J, LAM2L, Y2L):
    #     return LAM2L*LAM2L*Y2L*(Y2L-4)+4*J*(J+1)
    #
    # def C32L(J, LAM2L, Y2L):
    #     return LAM2L*LAM2L*Y2L*(Y2L-4)*(J-LAM2L)*(J+LAM2L+1)+2*(2*J+1)*(J-LAM2L+1)*(J+1)*(J+LAM2L+1)






from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99
# from a_WState import WState
# import moldb as db
from .a_WFileMolDB import *
import os
from ..a_XFileMainWindow import *


__all__ = ["XFileMolDB"]


class XFileMolDB(XFileMainWindow):
    def __init__(self, parent=None, fileobj=None):
        XFileMainWindow.__init__(self, parent)

        import f311.filetypes as ft


        # # Synchronized sequences
        _VVV = ft.FileMolDB.description
        self.tab_texts[0] =  "{} (Alt+&1)".format(_VVV)
        self.tabWidget.setTabText(0, self.tab_texts[0])
        self.save_as_texts[0] = "Save %s as..." % _VVV
        self.open_texts[0] = "Load %s" % _VVV
        self.clss[0] = ft.FileMolDB
        self.clsss[0] = (ft.FileMolDB,)
        self.wilds[0] = "*.sqlite"


        lv = self.keep_ref(QVBoxLayout(self.gotting))
        me = self.moldb_editor = WFileMolDB(self)
        lv.addWidget(me)
        me.changed.connect(self._on_changed)
        self.editors[0] = me

        self.setWindowTitle("Molecular information database editor")

        if fileobj is not None:
            self.load(fileobj)

    def wants_auto(self):
        idx = self.w_source.index
        filename = None
        if idx == 0:
            lines = self.w_hitran.data
            if lines:
                filename = "{}.dat".format(lines["header"]["table_name"])

        if filename is None:
            # Default
            filename = a99.new_filename("mol", "dat")

        self.w_out.value = filename


    def on_fill_missing(self):
        self.w_mol.None_to_zero()
        self.w_state.None_to_zero()

    # def mol_id_changed(self):
    #     id_ = self.w_mol.w_mol.id
    #     row = self.w_mol.w_mol.row
    #     self.w_state.set_id_molecule(id_)
    #     s = "States (no molecule selected)" if not row else "Select a State for molecule '{}'".format(row["formula"])
    #     self.title_state.setText(a99.format_title0(s))


    # # Override
    #   ========

    def _on_changed(self):
        """Overriden to commit automatically. "(changed)" will not appear

        Changes should be committed by the method that executed queries, but it is so easy to
        reinforce this... just in case.
        """
        index = self._get_tab_index()
        if index == 0:
            self.moldb_editor.f.get_conn().commit()  # Just in case
        else:
            self.flags_changed[index] = True
            self._update_tab_texts()



#!/usr/bin/env python

"""Molecular lines file editor."""

import sys
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
import argparse
import logging
import f311.pyfant as pf
import a99
import f311.filetypes as ft
import f311.explorer as ex


if __name__ == "__main__":
    a99.logging_level = logging.INFO
    a99.flag_log_file = True

    parser = argparse.ArgumentParser(
    description=__doc__,
    formatter_class=a99.SmartFormatter
    )
    parser.add_argument('fn', type=str, help='molecules file name',
                        default=ft.FileMolecules.default_filename, nargs='?')
    args = parser.parse_args()

    m = ft.FileMolecules()
    m.load(args.fn)
    app = a99.get_QApplication([])
    form = ex.XFileMolecules()
    form.show()
    form.load(m)
    sys.exit(app.exec_())














#!/usr/bin/env python
# coding: utf8
"""
Script to put together the file moldb.sqlite

**Note** This script is protected, but is very important to put to put together a lot of information
needed for molecular lines conversion (``convmol.py``)
"""

from f311 import filetypes as ft
import f311.convmol as cm
import a99
from collections import OrderedDict
import sqlite3
import os
import sys
from f311 import convmol as cm
from f311 import pyfant as pf


def insert_molecules(moldb):
    """Inserts data into the 'molecule' table"""
    assert isinstance(moldb, ft.FileMolDB)

    # Molecules present in PFANT molecules.dat
    MOLECULES = [("MgH", "Magnesium monohydride"),
                 ("C2", "Dicarbon"),
                 ("CN", "Cyano radical"),
                 ("CH", "Methylidyne"),
                 ("NH", "Imidogen"),
                 ("CO", "Carbon monoxide"),
                 ("OH", "Hydroxyl radical"),
                 ("FeH", "Iron hydride"),
                 ("TiO", "Titanium oxide")]

    conn = moldb.get_conn()
    conn.executemany("insert into molecule (formula, name) values (?,?)", MOLECULES)
    conn.commit()


def insert_pfantmol_data(moldb):
    """Inserts molecular header information 'PFANT/data/common/molecules.dat'"""
    assert isinstance(moldb, ft.FileMolDB)

    conn = moldb.get_conn()

    # Uses PFANT/data/common/molecules.dat to retrieve "fe", "do", "am", etc.
    filemol = ft.FileMolecules()
    filemol.load(pf.get_pfant_data_path("common", "molecules.dat"))
    # bysym = dict([(tuple(m.symbols), m) for m in filemol])

    for m in filemol:
        id_molecule = conn.execute("select id from molecule where formula = ?",
                                   (moldb.symbols_to_formula(m.symbols),)).fetchone()["id"]

        conn.execute("insert into pfantmol "
            "(id_molecule, description, fe, do, am, bm, ua, ub, te, cro, s) "
            "values (?,?,?,?,?,?,?,?,?,?,?)",
            (id_molecule, m.description, m.fe, m.do, m.am, m.bm, m.ua, m.ub, m.te,
            m.cro, m.s))
    conn.commit()


def insert_nist_data(moldb):
    """Tries to download data from NIST Web Book online"""
    assert isinstance(moldb, ft.FileMolDB)

    conn = moldb.get_conn()

    for row in conn.execute("select id, formula from molecule order by id").fetchall():
        id_molecule, formula = row["id"], row["formula"]
        a99.get_python_logger().info("Molecule '{}'...".format(formula))
        try:
            data, _, _ = cm.get_nist_webbook_constants(formula)

            for state in data:
                # **Note** assumes that the columns in data match the
                # (number of columns in the state table - 2) and their order
                conn.execute("insert into state values (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)",
                             [None, id_molecule] + state + [""])

            conn.commit()

        except:
            a99.get_python_logger().exception("Failed for molecule '{}'".format(formula))
    conn.commit()


def load_list_file(filename):
    """Loads a file containing (vl, v2l, FranckCondonFactor)

    ===begin exert===
    00 00 sj0000.dat .8638228E+00
    00 01 sj0001.dat .8712862E-01
    00 02 sj0002.dat .3144656E-01
    ===end exert===
    """


    fcfs = OrderedDict()
    with open(filename, "r") as h:
        for line in h:
            vl, v2l, _, fcf = line.strip().split(" ")
            vl = int(vl)
            v2l = int(v2l)
            fcf = float(fcf)
            fcfs[(vl, v2l)] = fcf
    return fcfs



# Electronic systems
#
# # TODO calculate FCFs for Tio, FeH, C2, CO
SYSTEMS_MAP = (
("CH", "cha.out", "A", 2, 2, "X", 2, 1, "Source: 'ATMOS/wrk4/bruno/Mole/Fc'"),  # A2Delta - X2Pi
("CH", "chb.out", "B", 2, 2, "X", 2, 1, "Source: 'ATMOS/wrk4/bruno/Mole/Fc'"),
("CH", "chc.out", "C", 2, 0, "X", 2, 1, "Source: 'ATMOS/wrk4/bruno/Mole/Fc'"),
("CN", "cnb.out", "B", 2, 0, "X", 2, 0, "Source: 'ATMOS/wrk4/bruno/Mole/Fc'"),
("NH", "nha.out", "A", 3, 1, "X", 3, 0, "Source: 'ATMOS/wrk4/bruno/Mole/Fc'"),  # A3Pi - X3Sigma
("OH", "oha.out", "A", 2, 0, "X", 2, 1, "Source: 'ATMOS/wrk4/bruno/Mole/Fc'"),  # A2Sigma - X2Pi
("MgH", "mgha.out", "A", 2, 1, "X", 2, 0, "Source: 'ATMOS/wrk4/bruno/Mole/Fc'"),
("MgH", "mghb.out", "B", 2, 0, "X", 2, 0, "Source: 'ATMOS/wrk4/bruno/Mole/Fc'"),
)

def my_info(s):
    a99.get_python_logger().info("[build-moldb] {}".format(s))

def insert_systems_and_franck_condon_factors(moldb):
    conn = moldb.get_conn()
    assert isinstance(conn, sqlite3.Connection)

    for formula, filename, from_label, from_mult, from_spdf, to_label, to_mult, to_spdf, comments in SYSTEMS_MAP:
        id_molecule = conn.execute("select id from molecule where formula = ?",
                                   (formula,)).fetchone()["id"]
        cursor = conn.execute("""insert into system (id_molecule, from_label, from_mult, from_spdf,
           to_label, to_mult, to_spdf, comments) values (?, ?, ?, ?, ?, ?, ?, ?)""",
           (id_molecule, from_label, from_mult, from_spdf, to_label, to_mult, to_spdf, comments))
        id_system = cursor.lastrowid

        if filename is not None:
            # Can handle two different file formats
            try:
                a = ft.FileFCF()
                a.load(filename)
                fcf_dict = a.fcfs
            except:
                fcf_dict = load_list_file(filename)
            for (vl, v2l), fcf in fcf_dict.items():
                conn.execute("insert into fcf (id_system, vl, v2l, value) values (?,?,?,?)",
                             (id_system, vl, v2l, fcf))
    conn.commit()

if __name__ == "__main__":

    filename = ft.FileMolDB.default_filename

    if os.path.isfile(filename):
        if a99.yesno("File '{}' already exists, get rid of it and continue".format(filename), True):
            os.unlink(filename)
        else:
            sys.exit()

    yn = a99.yesno("Will now create file '{}'. Continue".format(filename), True)

    if not yn:
        sys.exit()

    f = ft.FileMolDB()
    f.filename = filename
    my_info("Creating schema...")
    f.create_schema()
    my_info("New filename: '{}'".format(f.filename))
    my_info("Inserting molecules...")
    insert_molecules(f)
    my_info("Inserting Franck-Condon Factors from Bruno Castilho's work...")
    insert_systems_and_franck_condon_factors(f)
    my_info("Inserting molecule header information from '{}'...".
          format(pf.get_pfant_data_path("common", "molecules.dat")))
    insert_pfantmol_data(f)
    my_info("Inserting data from NIST Chemistry Web Book online...")
    insert_nist_data(f)
    # populate_moldb(f)
    # insert_franck_condon_factors(f)




from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99
from .a_WDBMolecule import WDBMolecule
from .a_WDBState import WDBState
from .a_WDBPFANTMol import *
from .a_WDBSystemFCF import *
import os


__all__ = ["WFileMolDB"]


class WFileMolDB(a99.WBase):

    @property
    def f(self):
        return self._f

    # @f.setter
    # def f(self, x):
    #     self._f = x
    #     self.w_mol.f = x
    #     self.w_state.f = x

    def __init__(self, *args):
        a99.WBase.__init__(self, *args)

        self._f = None
        self.flag_valid = True  # To keep XFileMainWindow happy

        # # Main layout & splitter
        lmain = self.keep_ref(QVBoxLayout(self))
        a99.set_margin(lmain, 0)
        sp = self.splitter_bidon = QSplitter(Qt.Horizontal)

        # ## Line showing the File Name
        l1 = self.keep_ref(QHBoxLayout())
        lmain.addLayout(l1)
        a99.set_margin(l1, 0)
        v = self.keep_ref(QLabel("<b>File:<b>"))
        v.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed))
        l1.addWidget(v)
        w = self.label_fn = QLabel()
        w.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed))
        l1.addWidget(w)
        l1.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))


        # ## First widget of splitter: molecules
        w0 = self.keep_ref(QWidget())
        l0 = QVBoxLayout(w0)
        a99.set_margin(l0, 2)
        l0.setSpacing(2)

        a = self.title_mol = QLabel(a99.format_title0("Molecules (Alt+&4)"))
        l0.addWidget(a)

        w = self.keep_ref(QLineEdit())
        l0.addWidget(w)

        w = self.w_mol = WDBMolecule(self.parent_form)
        # **Note** Cannot set buddy to w itself because it has a Qt.NoFocus policy
        a.setBuddy(w.tableWidget)
        w.id_changed.connect(self.mol_id_changed)
        w.changed.connect(self.changed)
        l0.addWidget(w)

        # ## Second widget of splitter: tab widget containing the rest
        w1 = self.tabWidget = QTabWidget(self)

        # ### First tab: molecule headers frmo PFANT molecular lines file
        w = self.w_pfantmol = WDBPFANTMol(self.parent_form)
        w.changed.connect(self.changed)
        w1.addTab(self.w_pfantmol, "PFANT molecules (Alt+&P)")
        # a99.set_margin(w10, 0)

        # ### Second tab: systems and Franck-Condon factors
        w = self.w_system = WDBSystemFCF(self.parent_form)
        w.changed.connect(self.changed)
        w1.addTab(self.w_system, "Electronic systems (Alt+&E)")

        # ### Thirs tab: NIST Chemistry Web Book data
        w = self.w_state = WDBState(self.parent_form)
        w.changed.connect(self.changed)
        w1.addTab(self.w_state, "States from NIST (Alt+&S)")

        sp.addWidget(w0)
        sp.addWidget(w1)
        # sp.setStretchFactor(0, 1)
        # sp.setStretchFactor(1, 2)

        lmain.addWidget(sp)

        # # Final adjustments
        a99.nerdify(self)

    def load(self, x):
        self._f = x
        self.w_mol.f = x
        self.w_state.f = x
        self.w_pfantmol.f = x
        self.w_system.f = x
        self.update_gui_label_fn()


    def mol_id_changed(self):
        id_ = self.w_mol.id
        row = self.w_mol.row
        self.w_pfantmol.set_id_molecule(id_)
        self.w_state.set_id_molecule(id_)
        self.w_system.set_id_molecule(id_)


    # # Override
    #   ========

    def update_gui_label_fn(self):
        if not self._f:
            text = "(not loaded)"
        elif self._f.filename:
            text = os.path.relpath(self._f.filename, ".")
        else:
            text = "(filename not set)"
        self.label_fn.setText(text)



#!/usr/bin/env python

"""Spectrum List Editor"""

import sys
import argparse
import a99
import f311.filetypes as ft
import f311.explorer as ex
import logging
import f311.explorer as ex
import f311.filetypes as ft

a99.logging_level = logging.INFO
a99.flag_log_file = True


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
    description=__doc__,
    formatter_class=a99.SmartFormatter
    )
    parser.add_argument('fn', type=str, help="file name, supports '%s' only at the moment" %
                                             (ft.FileSpectrumList.description,), nargs='?')
    args = parser.parse_args()

    app = a99.get_QApplication([])
    form = ex.XFileSpectrumList()

    if args.fn is not None:
        form.load_filename(args.fn)

    form.show()
    sys.exit(app.exec_())



from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99
from .a_WDBState import WDBState
import copy

__all__ = ["WMolecularConstants"]


class WMolecularConstants(a99.WBase):
    """
    Widget for the user to type or obtain many molecular constants
    """

    @property
    def f(self):
        """Object representing the file being edited (possibly a DataFile object)"""
        return self._f

    @f.setter
    def f(self, f):
        self.load(f)

    @property
    def constants(self):
        """Returns a dictionary that combines several records (see details):

        - one record from table 'molecule'
        - one record from table 'system'
        - one record from table 'pfantmol'
        - two records from table 'state', for state' and state''
            * keys for state' values start with "statel_"
            * keys for state'' values start with "state2l_"
        """
        ret = {}
        for fieldname in self._fieldnames:
            ret[fieldname] = self[fieldname]
        ret["name"] = self.name
        ret["formula"] = self.formula
        return ret

    @property
    def name(self):
        """Returns molecule 'name' (field from 'molecule' table)"""
        ret, id_ = None, self._get_id_molecule()
        if id_ is not None:
            ret = self._f.get_conn().execute("select name from molecule where id = ?", (id_,)).fetchone()["name"]
        return ret

    @property
    def formula(self):
        """Returns molecule 'formula' (field from 'molecule' table)"""
        ret, id_ = None, self._get_id_molecule()
        if id_ is not None:
            ret = self._f.get_conn().execute("select formula from molecule where id = ?", (id_,)).fetchone()["formula"]
        return ret

    @property
    def fcfs(self):
        """Returns a dictionary of Franck-Condon Factors (key is (vl, v2l)), or None"""
        return self._get_fcfs()

    @property
    def id_molecule(self):
        return self._get_id_molecule()

    # @property
    # def row(self):
    #     """Wraps WDBMolecule.row"""
    #     return self.w_mol.row

    def __init__(self, *args):
        a99.WBase.__init__(self, *args)

        self.flag_populating = False  # activated when populating table

        # # Internal state

        # FileMolDB object, I guess
        self._f = None

        # Fields of interest from table 'pfantmol'
        self._fieldnames_pfantmol = ["fe", "do", "am", "bm", "ua", "ub", "te", "cro", "s", ]
        # Fields of interest from table 'state'
        self._fieldnames_state = ["omega_e", "omega_ex_e", "omega_ey_e", "B_e", "alpha_e", "D_e",
                                 "beta_e", "A"]
        # Fields of interest from table 'system'
        self._fieldnames_system = ["from_label", "from_spdf", "to_label", "to_spdf"]
        self._flag_built_edits = False
        for fn in self._fieldnames_state:
            assert fn not in self._fieldnames_pfantmol
            assert fn not in self._fieldnames_system
        for fn in self._fieldnames_pfantmol:
            assert fn not in self._fieldnames_system
        self._fieldnames = []  # will be filled in later copy.copy(self._fieldnames_pfantmol)
        # self._fieldnames.extend(self._fieldnames_state)
        # self._fieldnames.extend(self._fieldnames_system)

        # dictionary {(field name): (edit object), }
        # (will be populated later below together with edit widgets creation)
        self._edit_map = {}
        self._edit_map_statel = {}
        self._edit_map_state2l = {}
        self._edit_map_pfantmol = {}
        self._edit_map_system = {}

        # id from table 'molecule'. In sync with combobox_select_molecule items
        self._ids_molecule = []
        # id from table 'pfantmol'. In sync with combobox_select_pfantmol
        self._ids_pfantmol = []
        # id from table 'state'. In sync with combobox_select_statel and combobox_state2l
        self._ids_state = []
        # id from table 'system'. In sync with combobox_select_system
        self._ids_system = []

        # # GUI design

        l = self.layout_main = QVBoxLayout(self)
        a99.set_margin(l, 2)
        l.setSpacing(6)

        # ## Toolbar at top
        bb = self.button_zeros = QPushButton("Fill empty fields with zeros")
        bb.clicked.connect(self.None_to_zero)
        l.addWidget(bb)

        # ## Select molecule combobox
        l0 = self.layout_select_molecule = QHBoxLayout()
        l.addLayout(l0)
        l0.setSpacing(3)
        la = self.label_select_molecule = QLabel("<b>Molecule</b>")
        la.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed))
        l0.addWidget(la)
        cb = self.combobox_select_molecule = QComboBox()
        cb.currentIndexChanged.connect(self.combobox_select_molecule_currentIndexChanged)
        l0.addWidget(cb)

        # ## Frame with the PFANT molecule-wide fields

        fr = self.frame_pfantmol = a99.get_frame()
        l.addWidget(fr)
        l1 = self.layout_frame_pfantmol = QVBoxLayout(fr)

        # ### Select pfantmol combobox
        l2 = self.layout_select_pfantmol = QHBoxLayout()
        l1.addLayout(l2)
        l2.setSpacing(3)
        la = self.label_select_pfantmol = QLabel("<b>PFANT molecule</b>")
        la.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed))
        l2.addWidget(la)
        cb = self.combobox_select_pfantmol = QComboBox()
        cb.currentIndexChanged.connect(self.combobox_select_pfantmol_currentIndexChanged)
        l2.addWidget(cb)

        # ### PFANT Molecular constants edit fields

        lg = self.layout_grid_pfantmol = QGridLayout()
        l1.addLayout(lg)

        # ## Frame with the combobox to select the "system"

        fr = self.frame_system = a99.get_frame()
        l.addWidget(fr)
        l1 = self.layout_frame_system = QVBoxLayout(fr)

        # ### Select system combobox
        l2 = self.layout_select_system = QHBoxLayout()
        l1.addLayout(l2)
        l2.setSpacing(3)
        la = self.label_select_system = QLabel("<b>Electronic system</b>")
        la.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed))
        l2.addWidget(la)
        cb = self.combobox_select_system = QComboBox()
        cb.currentIndexChanged.connect(self.combobox_select_system_currentIndexChanged)
        l2.addWidget(cb)

        la = self.label_fcf = QLabel()
        la.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed))
        l1.addWidget(la)

        # ### System constants edit fields

        lg = self.layout_grid_system = QGridLayout()
        l1.addLayout(lg)

        # #### Frame for Diatomic molecular constants for statel

        fr = self.frame_statel = a99.get_frame()
        l1.addWidget(fr)
        l3 = self.layout_frame_statel = QVBoxLayout(fr)

        # ##### Select statel combobox
        l4 = self.layout_select_statel = QHBoxLayout()
        l3.addLayout(l4)
        l4.setSpacing(3)
        la = self.label_select_statel = QLabel("<b>Diatomic molecular constants for state'</b>")
        la.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed))
        l4.addWidget(la)
        cb = self.combobox_select_statel = QComboBox()
        cb.currentIndexChanged.connect(self.combobox_select_statel_currentIndexChanged)
        l4.addWidget(cb)

        # ##### State' edit fields

        lg = self.layout_grid_statel = QGridLayout()
        l3.addLayout(lg)





        # #### Frame for Diatomic molecular constants for state2l

        fr = self.frame_state2l = a99.get_frame()
        l1.addWidget(fr)
        l3 = self.layout_frame_state2l = QVBoxLayout(fr)

        # ##### Select state2l combobox
        l4 = self.layout_select_state2l = QHBoxLayout()
        l3.addLayout(l4)
        l4.setSpacing(3)
        la = self.label_select_state2l = QLabel("<b>Diatomic molecular constants for state''</b>")
        la.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed))
        l4.addWidget(la)
        cb = self.combobox_select_state2l = QComboBox()
        cb.currentIndexChanged.connect(self.combobox_select_state2l_currentIndexChanged)
        l4.addWidget(cb)

        # ##### State' edit fields

        lg = self.layout_grid_state2l = QGridLayout()
        l3.addLayout(lg)









        a99.nerdify(self)

    def __getitem__(self, fieldname):
        """Allows dict-like access of molecular constants of interest. Returns float value or None"""
        if fieldname in self._fieldnames:
            text = self._edit_map[fieldname].text()
            if "_label" in fieldname:
                # Fields with "_label" in their names are treated as strings,
                # otherwise they are considered numeric
                return text.upper()
            else:
                try:
                    return float(text)
                except ValueError:
                    return None


    def load(self, f):
        """Loads a FileMolDB object"""
        import f311.filetypes as ft
        assert isinstance(f, ft.FileMolDB)

        self._f = f
        if f is not None:
            self._populate()


    def None_to_zero(self):
        """Fills missing values with zeros"""
        for edit in self._edit_map.values():
            if len(edit.text().strip()) == 0:
                edit.setText("0")

    ################################################################################################
    # # Slots

    def combobox_select_molecule_currentIndexChanged(self):
        self._populate_sub_comboboxes()

    def combobox_select_pfantmol_currentIndexChanged(self):
        self._fill_edits_pfantmol()

    def combobox_select_statel_currentIndexChanged(self):
        self._fill_edits_statel()

    def combobox_select_state2l_currentIndexChanged(self):
        self._fill_edits_state2l()

    def combobox_select_system_currentIndexChanged(self):
        self._fill_edits_system()
        self._update_label_fcf()


    ################################################################################################
    # # Internal function

    def _get_id_molecule(self):
        if len(self._ids_molecule) > 0:
            return self._ids_molecule[self.combobox_select_molecule.currentIndex()]
        return None

    def _get_id_pfantmol(self):
        idx = self.combobox_select_pfantmol.currentIndex()
        if idx > 0:
            return self._ids_pfantmol[idx-1]
        return None

    def _get_id_statel(self):
        idx = self.combobox_select_statel.currentIndex()
        if idx > 0:
            return self._ids_state[idx-1]
        return None

    def _get_id_state2l(self):
        idx = self.combobox_select_state2l.currentIndex()
        if idx > 0:
            return self._ids_state[idx-1]
        return None

    def _get_id_system(self):
        if len(self._ids_system) > 0:
            return self._ids_system[self.combobox_select_system.currentIndex()]
        return None

    def _get_fcfs(self):
        ret, id_ = None, self._get_id_system()
        if id_ is not None:
            _ret = self._f.query_fcf(id_system=id_).fetchall()
            ret = {}
            for r in _ret:
                ret[(r["vl"], r["v2l"])] = r["value"]
        return ret

    def _populate(self):
        if not self._flag_built_edits:
            self._build_edits()
        self._populate_combobox_select_molecule()
        self._populate_sub_comboboxes()

    def _populate_sub_comboboxes(self):
        self._populate_combobox_select_pfantmol()
        self._populate_combobox_select_state()
        self._populate_combobox_select_system()

    def _populate_combobox_select_molecule(self):
        cb = self.combobox_select_molecule
        cb.clear()
        self._ids_molecule = []
        cursor = self._f.query_molecule()
        for row in cursor:
            cb.addItem("{:10} {}".format(row["formula"], row["name"]))
            self._ids_molecule.append(row["id"])

    def _populate_combobox_select_pfantmol(self):
        cb = self.combobox_select_pfantmol
        cb.clear()
        self._ids_pfantmol = []
        data = self._f.query_pfantmol(id_molecule=self._get_id_molecule()).fetchall()
        cb.addItem("(select to fill information below)" if len(data) > 0 else "(no data)")
        for row in data:
            cb.addItem("{}".format(row["description"]))
            self._ids_pfantmol.append(row["id"])
        self._fill_edits_pfantmol()

    def _populate_combobox_select_state(self):
        self._ids_state = []
        data = self._f.query_state(id_molecule=self._get_id_molecule()).fetchall()
        cb = self.combobox_select_statel
        cb.clear()
        cb.addItem("(select to fill information below)" if len(data) > 0 else "(no data)")
        for row in data:
            cb.addItem("{}".format(row["State"]))
            self._ids_state.append(row["id"])

        cb = self.combobox_select_state2l
        cb.clear()
        cb.addItem("(select to fill information below)" if len(data) > 0 else "(no data)")
        for row in data:
            cb.addItem("{}".format(row["State"]))

        self._fill_edits_statel()
        self._fill_edits_state2l()

    def _populate_combobox_select_system(self):
        cb = self.combobox_select_system
        cb.clear()
        self._ids_system = []
        data = self._f.query_system(id_molecule=self._get_id_molecule()).fetchall()
        if len(data) == 0:
            cb.addItem("(no data)")
        for row in data:
            cb.addItem(self._f.get_system_short(row))
            self._ids_system.append(row["id"])
        self._fill_edits_system()
        self._update_label_fcf()


    def __build_edits_generic(self, lg, fn, nc, ti, em):
        """Populates a grid layout with edit fields

        Args:
            lg: QGridLayout
            fn: field names
            nc: number of columns
            ti: dict-like object with information about fields
            em: edit map
        """
        n = len(fn)
        for j in range(nc):
            # ### One grid layout for each column of fields
            ii = range(j, n, nc)
            for i in range(len(ii)):
                fieldname = fn[ii[i]]
                info = ti[fieldname]
                caption = info["caption"] or fieldname
                a = QLabel(caption)
                e = QLineEdit("")
                tooltip = info["tooltip"]
                if tooltip:
                    a.setToolTip(tooltip)
                    e.setToolTip(tooltip)
                self._edit_map[fieldname] = e
                em[fieldname] = e
                lg.addWidget(a, i, j * 2)
                lg.addWidget(e, i, j * 2 + 1)

    def _build_edits(self):
        # pfantmol fields
        nr, nc, n = 3, 3, len(self._fieldnames_pfantmol)
        ti = self._f.get_table_info("pfantmol")
        lg = self.layout_grid_pfantmol
        for j in range(nc):
            # ### One grid layout for each column of fields
            ii = range(j, n, nc)
            for i in range(len(ii)):
                fieldname = self._fieldnames_pfantmol[ii[i]]
                info = ti[fieldname]
                caption = info["caption"] or fieldname
                a = QLabel(caption)
                e = QLineEdit("")
                tooltip = info["tooltip"]
                if tooltip:
                    a.setToolTip(tooltip)
                    e.setToolTip(tooltip)
                self._fieldnames.append(fieldname)
                self._edit_map[fieldname] = e
                self._edit_map_pfantmol[fieldname] = e
                lg.addWidget(a, i, j * 2)
                lg.addWidget(e, i, j * 2 + 1)

        # system fields
        nr, nc, n = 2, 2, len(self._fieldnames_system)
        ti = self._f.get_table_info("system")
        lg = self.layout_grid_system
        for j in range(nc):
            # ### One grid layout for each column of fields
            ii = range(j, n, nc)
            for i in range(len(ii)):
                fieldname = self._fieldnames_system[ii[i]]
                info = ti[fieldname]
                caption = info["caption"] or fieldname
                a = QLabel(caption)
                e = QLineEdit("")
                tooltip = info["tooltip"]
                if tooltip:
                    a.setToolTip(tooltip)
                    e.setToolTip(tooltip)
                self._fieldnames.append(fieldname)
                self._edit_map[fieldname] = e
                self._edit_map_system[fieldname] = e
                lg.addWidget(a, i, j * 2)
                lg.addWidget(e, i, j * 2 + 1)

        # statel fields
        nr, nc, n = 3, 3, len(self._fieldnames_state)
        ti = self._f.get_table_info("state")
        lg = self.layout_grid_statel
        for j in range(nc):
            # ### One grid layout for each column of fields
            ii = range(j, n, nc)
            for i in range(len(ii)):
                fieldname = self._fieldnames_state[ii[i]]
                info = ti[fieldname]
                caption = info["caption"] or fieldname
                a = QLabel(caption)
                e = QLineEdit("")
                tooltip = info["tooltip"]
                if tooltip:
                    a.setToolTip(tooltip)
                    e.setToolTip(tooltip)
                fieldname_all = "statel_" + fieldname
                self._fieldnames.append(fieldname_all)
                self._edit_map[fieldname_all] = e
                self._edit_map_statel[fieldname] = e
                lg.addWidget(a, i, j * 2)
                lg.addWidget(e, i, j * 2 + 1)

        # statel fields
        nr, nc, n = 3, 3, len(self._fieldnames_state)
        ti = self._f.get_table_info("state")
        lg = self.layout_grid_state2l
        for j in range(nc):
            # ### One grid layout for each column of fields
            ii = range(j, n, nc)
            for i in range(len(ii)):
                fieldname = self._fieldnames_state[ii[i]]
                info = ti[fieldname]
                caption = info["caption"] or fieldname
                a = QLabel(caption)
                e = QLineEdit("")
                tooltip = info["tooltip"]
                if tooltip:
                    a.setToolTip(tooltip)
                    e.setToolTip(tooltip)
                fieldname_all = "state2l_" + fieldname
                self._fieldnames.append(fieldname_all)
                self._edit_map[fieldname_all] = e
                self._edit_map_state2l[fieldname] = e
                lg.addWidget(a, i, j * 2)
                lg.addWidget(e, i, j * 2 + 1)


        self._flag_built_edits = True

    def _fill_edits_pfantmol(self):
        id_ = self._get_id_pfantmol()
        if id_ is not None:
            row = self._f.query_pfantmol(**{"pfantmol.id": id_}).fetchone()
            for fieldname, e in self._edit_map_pfantmol.items():
                v = row[fieldname]
                e.setText(str(v) if v is not None else "")

    def _fill_edits_statel(self):
        id_ = self._get_id_statel()
        if id_ is not None:
            row = self._f.query_state(**{"state.id": id_}).fetchone()
            for fieldname, e in self._edit_map_statel.items():
                v = row[fieldname]
                e.setText(str(v) if v is not None else "")

    def _fill_edits_state2l(self):
        id_ = self._get_id_state2l()
        if id_ is not None:
            row = self._f.query_state(**{"state.id": id_}).fetchone()
            for fieldname, e in self._edit_map_state2l.items():
                v = row[fieldname]
                e.setText(str(v) if v is not None else "")


    def _fill_edits_system(self):
        id_ = self._get_id_system()
        if id_ is not None:
            row = self._f.query_system(**{"system.id": id_}).fetchone()
            for fieldname, e in self._edit_map_system.items():
                v = row[fieldname]
                e.setText(str(v) if v is not None else "")

    def _update_label_fcf(self):
        n, fcfs = 0, self._get_fcfs()
        if fcfs is not None:
            n = len(fcfs)
        s = '<font color="{}">Franck-Condon Factors (FCFs) for {} vibrational transition{}</font>'.\
            format("blue" if n != 0 else "red", n, "" if n == 1 else "s")
        self.label_fcf.setText(s)

